/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA.
 */

#include "gtk-ppp.h"

struct _binarydescr {
	gchar *name;
	gchar *path;
	gboolean exists;
	gboolean suid;
};

typedef struct _binarydescr *binarydescr_t;

static void gtk_ppp_show_message(gchar *message)
{
	GtkWidget *dialog;

	dialog = gtk_message_dialog_new(NULL,
									GTK_DIALOG_MODAL,
									GTK_MESSAGE_ERROR,
									GTK_BUTTONS_CLOSE,
									message,
									NULL);

	gtk_dialog_run(GTK_DIALOG(dialog));
	gtk_widget_destroy(dialog);
}

static void gtk_ppp_binaries_description_table_destroy(gpointer data)
{
	binarydescr_t binarydescr;

	if (data == NULL) return;

	binarydescr = (binarydescr_t)data;

	if (binarydescr->name != NULL) g_free(binarydescr->name);
	if (binarydescr->path != NULL) g_free(binarydescr->path);
	g_free(binarydescr);
}

static GHashTable *gtk_ppp_discover_permissions(const gchar *firstbinary, ...)
{
	const gchar *sysdirs[] = {"/bin", "/sbin", "/usr/bin", "/usr/sbin", "/usr/local/bin", "/usr/local/sbin", "/opt/bin", NULL};
	va_list args;
	const gchar *usrdirsstr, *curbinary;
	gchar **usrdirs;
	GHashTable *binstab, *dirstab;
	GHashTableIter iter;
	gchar *execpath, *bindir;
	gint i;
	struct stat execstatbuf;
	binarydescr_t binarydescr;
	
	if (firstbinary == NULL) return NULL;
		
	dirstab = g_hash_table_new(g_str_hash, g_str_equal);
	
	/*User binary directories*/
	usrdirsstr = g_getenv("PATH");
	usrdirs = g_strsplit(usrdirsstr, ":", -1);
	i = 0;
	while (usrdirs[i] != NULL) {
		g_hash_table_insert(dirstab, usrdirs[i], usrdirs[i]);
		i++;
	}
	/*Administrator binary directories*/
	i = 0;
	while (sysdirs[i] != NULL) {
		if (!g_hash_table_contains(dirstab, sysdirs[i])) {
			g_hash_table_insert(dirstab, (gchar *)sysdirs[i], (gchar *)sysdirs[i]);
		}
		i++;
	}
	
	/*Search for specified binaries*/
	binstab = g_hash_table_new_full(g_str_hash, g_str_equal, NULL, gtk_ppp_binaries_description_table_destroy);
	
	va_start(args, firstbinary);
	curbinary = firstbinary;
	
	do {
		binarydescr = g_new0(struct _binarydescr, 1);
		binarydescr->name = g_strdup(curbinary);
		binarydescr->path = NULL;
		binarydescr->exists = FALSE;
		binarydescr->suid = FALSE;
		g_hash_table_iter_init(&iter, dirstab);
		while (g_hash_table_iter_next(&iter, (gpointer *)&bindir, NULL)) {
			execpath = g_build_filename(bindir, curbinary, NULL);
			if (stat(execpath, &execstatbuf) != -1) {
				binarydescr->path = g_strdup(bindir);
				binarydescr->exists = (gboolean)(execstatbuf.st_mode & S_IFREG);
				binarydescr->suid = (gboolean)(execstatbuf.st_mode & S_ISUID);
				break;
			}
			g_free(execpath);
		}
		g_hash_table_insert(binstab, binarydescr->name, binarydescr);
		curbinary = va_arg(args, gchar *);
	} while (curbinary != NULL);
	
	/*Free temporary data structures*/
	g_hash_table_destroy(dirstab);
	g_strfreev(usrdirs);
	
	return binstab;
}

static gboolean gtk_ppp_binary_exists(GHashTable *bintab, const gchar *binary)
{
	binarydescr_t binarydescr;

	if ((bintab == NULL) || (binary == NULL)) return FALSE;

	binarydescr = g_hash_table_lookup(bintab, binary);

	if (binarydescr == NULL) return FALSE;

	return binarydescr->exists;
}

static gboolean gtk_ppp_binary_suid(GHashTable *bintab, const gchar *binary)
{
	binarydescr_t binarydescr;

	if ((bintab == NULL) || (binary == NULL)) return FALSE;

	binarydescr = g_hash_table_lookup(bintab, binary);

	if (binarydescr == NULL) return FALSE;

	return binarydescr->suid;
}

static gchar *gtk_ppp_binary_path(GHashTable *bintab, const gchar *binary)
{
	binarydescr_t binarydescr;

	if ((bintab == NULL) || (binary == NULL)) return NULL;

	binarydescr = g_hash_table_lookup(bintab, binary);

	if (binarydescr == NULL) return NULL;

	return g_build_filename(binarydescr->path, binarydescr->name, NULL);
}

static gboolean gtk_ppp_test_enviroment(void)
{
	GHashTable *bintab;
	gboolean res;
	GError *error = NULL;
	gchar *argv[4];
	GPid pid;

	bintab = gtk_ppp_discover_permissions("wvdial", "pppd", "pkexec", NULL);

	if (bintab == NULL) {
		gtk_ppp_show_message(_("Something wrong with permissions. Please start application with root permissions."));
		return FALSE;
	}

	if (geteuid() == 0) {
		if (!gtk_ppp_binary_exists(bintab, "wvdial")) {
			gtk_ppp_show_message(_("Unable to find WVDIAL executable file. Please make sure you installed it properly."));
			g_hash_table_destroy(bintab);
			return FALSE;
		} else if (!gtk_ppp_binary_exists(bintab, "pppd")) {
			gtk_ppp_show_message(_("Unable to find PPPD executable file. Please make sure you installed it properly."));
			g_hash_table_destroy(bintab);
			return FALSE;
		}
		g_hash_table_destroy(bintab);
		return TRUE;
	}

	if (!((gtk_ppp_binary_suid(bintab, "wvdial")) && (gtk_ppp_binary_suid(bintab, "pppd")))) {
		if (!gtk_ppp_binary_exists(bintab, "wvdial")) {
			gtk_ppp_show_message(_("Unable to find WVDIAL executable file. Please make sure you installed it properly."));
		} else if (!gtk_ppp_binary_exists(bintab, "pppd")) {
			gtk_ppp_show_message(_("Unable to find PPPD executable file. Please make sure you installed it properly."));
		} else if ((!(gtk_ppp_binary_suid(bintab, "wvdial"))) || (!(gtk_ppp_binary_suid(bintab, "pppd")))) {
			#ifdef ENABLE_POLKIT
			if (gtk_ppp_binary_suid(bintab, "pkexec")) {
				argv[0] = gtk_ppp_binary_path(bintab, "pkexec");
				argv[1] = "--disable-internal-agent";
				argv[2] = PACKAGE_EXECUTABLE_PATH;
				argv[3] = NULL;

				res = g_spawn_async(NULL,
									argv,
									NULL,
									G_SPAWN_DO_NOT_REAP_CHILD,
									NULL,
									NULL,
									&pid,
									&error);

				if (res) {
					waitpid(0, NULL, 0);
				} else {
					gtk_ppp_show_message(_("Unable to restart GTK PPP with root permissions"));
					g_error_free(error);
				}

				g_free(argv[0]);
			} else
			#endif
			{
				if (!(gtk_ppp_binary_suid(bintab, "wvdial"))) {
					gtk_ppp_show_message(_("WVDIAL executable file has no SUID bit. Please start application with root permissions."));
				} else if (!(gtk_ppp_binary_suid(bintab, "pppd"))) {
					gtk_ppp_show_message(_("PPPD executable file has no SUID bit. Please start application with root permissions."));
				}
			}

		}

		g_hash_table_destroy(bintab);

		return FALSE;
	}

	g_hash_table_destroy(bintab);

	return TRUE;
}

static GtkWidget *gtk_ppp_create_main_window(GtkApplication *application)
{
	gchar *uifile;
	gchar *errormsg;
	GError *error = NULL;
	gchar *icon;

	/* init GTKBuilder */
	uifile = g_build_filename(PACKAGE_DATA_DIR, "ui", "gtk-ppp.glade", NULL);
	gtk_ppp.builder = gtk_builder_new();
	if (!gtk_builder_add_from_file(gtk_ppp.builder, uifile, &error)) {
		errormsg = g_strdup_printf(_("Error loading UI file: %s"), error->message);
		gtk_ppp_show_message(errormsg);
		g_error_free(error);
		g_free(errormsg);
		g_free(uifile);
		return NULL;
	}
	g_free(uifile);

	/* app icon path variable */
	icon = g_build_filename(PACKAGE_APPICON_DIR, "gtk-ppp.png", NULL);

	/* gtk ppp window */
	gtk_ppp.gtk_ppp.window = GTK_WIDGET(gtk_builder_get_object(gtk_ppp.builder, "gtk_ppp"));
	gtk_window_set_icon_from_file (GTK_WINDOW (gtk_ppp.gtk_ppp.window), icon, NULL);
	gtk_ppp.gtk_ppp.username = GTK_WIDGET(gtk_builder_get_object(gtk_ppp.builder, "username"));
	gtk_ppp.gtk_ppp.password = GTK_WIDGET(gtk_builder_get_object(gtk_ppp.builder, "password"));
	gtk_ppp.gtk_ppp.store_pwd = GTK_WIDGET(gtk_builder_get_object(gtk_ppp.builder, "store_pwd"));
	gtk_ppp.gtk_ppp.phone = GTK_WIDGET(gtk_builder_get_object(gtk_ppp.builder, "phone"));
	gtk_ppp.gtk_ppp.phoneentry = GTK_WIDGET(gtk_builder_get_object(gtk_ppp.builder, "phone-entry"));
	/* when enter is pressed in phone combo, activate default button (`connect`) */
	gtk_entry_set_activates_default(GTK_ENTRY(gtk_ppp.gtk_ppp.phoneentry), TRUE);
	/* setup window */
	/* modem tab */
	gtk_ppp.setup.window = GTK_WIDGET(gtk_builder_get_object(gtk_ppp.builder, "setup"));
	gtk_window_set_icon_from_file (GTK_WINDOW (gtk_ppp.setup.window), icon, NULL);
	gtk_ppp.setup.device = GTK_WIDGET(gtk_builder_get_object(gtk_ppp.builder, "device"));
	gtk_ppp.setup.deviceentry = GTK_WIDGET(gtk_builder_get_object(gtk_ppp.builder, "device-entry"));
	gtk_ppp.setup.type = GTK_WIDGET(gtk_builder_get_object(gtk_ppp.builder, "type"));
	gtk_ppp.setup.speed = GTK_WIDGET(gtk_builder_get_object(gtk_ppp.builder, "speed"));
	gtk_ppp.setup.speedentry = GTK_WIDGET(gtk_builder_get_object(gtk_ppp.builder, "speed-entry"));
	gtk_ppp.setup.tone = GTK_WIDGET(gtk_builder_get_object(gtk_ppp.builder, "tone"));
	gtk_ppp.setup.pulse = GTK_WIDGET(gtk_builder_get_object(gtk_ppp.builder, "pulse"));
	gtk_ppp.setup.volume = GTK_WIDGET(gtk_builder_get_object(gtk_ppp.builder, "volume"));
	gtk_ppp.setup.phone_nums = GTK_WIDGET(gtk_builder_get_object(gtk_ppp.builder, "phone_nums"));
	gtk_ppp.setup.init_strs = GTK_WIDGET(gtk_builder_get_object(gtk_ppp.builder, "init_strs"));
	gtk_ppp.setup.dial_prefix = GTK_WIDGET(gtk_builder_get_object(gtk_ppp.builder, "dial_prefix"));
	gtk_ppp.setup.spin_dial_attempts = GTK_WIDGET(gtk_builder_get_object(gtk_ppp.builder, "spin_dial_attempts"));
	gtk_ppp.setup.label_dial_attempts = GTK_WIDGET(gtk_builder_get_object(gtk_ppp.builder, "label_dial_attempts"));
	gtk_ppp.setup.dial_tone = GTK_WIDGET(gtk_builder_get_object(gtk_ppp.builder, "dial_tone"));
	/* networking tab */
	gtk_ppp.setup.static_ip = GTK_WIDGET(gtk_builder_get_object(gtk_ppp.builder, "static_ip"));
	gtk_ppp.setup.dynamic_ip = GTK_WIDGET(gtk_builder_get_object(gtk_ppp.builder, "dynamic_ip"));
	gtk_ppp.setup.label_ip = GTK_WIDGET(gtk_builder_get_object(gtk_ppp.builder, "label_ip"));
	gtk_ppp.setup.entry_ip = GTK_WIDGET(gtk_builder_get_object(gtk_ppp.builder, "entry_ip"));
	gtk_ppp.setup.domain = GTK_WIDGET(gtk_builder_get_object(gtk_ppp.builder, "domain"));
	gtk_ppp.setup.auto_dns = GTK_WIDGET(gtk_builder_get_object(gtk_ppp.builder, "auto_dns"));
	gtk_ppp.setup.manual_dns = GTK_WIDGET(gtk_builder_get_object(gtk_ppp.builder, "manual_dns"));
	gtk_ppp.setup.list_dns = GTK_WIDGET(gtk_builder_get_object(gtk_ppp.builder, "list_dns"));
	/* options tab */
	gtk_ppp.setup.minimize = GTK_WIDGET(gtk_builder_get_object(gtk_ppp.builder, "minimize"));
	gtk_ppp.setup.dock = GTK_WIDGET(gtk_builder_get_object(gtk_ppp.builder, "dock"));
	gtk_ppp.setup.auto_reconnect = GTK_WIDGET(gtk_builder_get_object(gtk_ppp.builder, "auto_reconnect"));
	gtk_ppp.setup.abort_on_busy = GTK_WIDGET(gtk_builder_get_object(gtk_ppp.builder, "abort_on_busy"));
	gtk_ppp.setup.abort_on_no_dialtone = GTK_WIDGET(gtk_builder_get_object(gtk_ppp.builder, "abort_on_no_dialtone"));
	gtk_ppp.setup.check_carrier = GTK_WIDGET(gtk_builder_get_object(gtk_ppp.builder, "check_carrier"));
	gtk_ppp.setup.check_def_route = GTK_WIDGET(gtk_builder_get_object(gtk_ppp.builder, "check_def_route"));
	gtk_ppp.setup.stupid_mode = GTK_WIDGET(gtk_builder_get_object(gtk_ppp.builder, "stupid_mode"));
	gtk_ppp.setup.custom_reply = GTK_WIDGET(gtk_builder_get_object(gtk_ppp.builder, "custom_reply"));
	gtk_ppp.setup.label_reply = GTK_WIDGET(gtk_builder_get_object(gtk_ppp.builder, "label_reply"));
	gtk_ppp.setup.entry_reply = GTK_WIDGET(gtk_builder_get_object(gtk_ppp.builder, "entry_reply"));
	gtk_ppp.setup.spin_idle = GTK_WIDGET(gtk_builder_get_object(gtk_ppp.builder, "spin_idle"));
	gtk_ppp.setup.label_idle = GTK_WIDGET(gtk_builder_get_object(gtk_ppp.builder, "label_idle"));
	/* connecting window */
	gtk_ppp.connecting.window = GTK_WIDGET(gtk_builder_get_object(gtk_ppp.builder, "connecting"));
	gtk_window_set_icon_from_file (GTK_WINDOW (gtk_ppp.connecting.window), icon, NULL);
	gtk_ppp.connecting.msgs = GTK_WIDGET(gtk_builder_get_object(gtk_ppp.builder, "msgs"));
	/* log window */
	gtk_ppp.log.window = GTK_WIDGET(gtk_builder_get_object(gtk_ppp.builder, "log"));
	gtk_window_set_icon_from_file (GTK_WINDOW (gtk_ppp.log.window), icon, NULL);
	gtk_ppp.log.text_log = GTK_WIDGET(gtk_builder_get_object(gtk_ppp.builder, "text_log"));
	/* connected window */
	gtk_ppp.connected.window = GTK_WIDGET(gtk_builder_get_object(gtk_ppp.builder, "connected"));
	gtk_window_set_icon_from_file (GTK_WINDOW (gtk_ppp.connected.window), icon, NULL);
	gtk_ppp.connected.time = GTK_WIDGET(gtk_builder_get_object(gtk_ppp.builder, "time"));
	/* details window */
	gtk_ppp.details.window = GTK_WIDGET(gtk_builder_get_object(gtk_ppp.builder, "details"));
	gtk_window_set_icon_from_file (GTK_WINDOW (gtk_ppp.details.window), icon, NULL);
	gtk_ppp.details.interface = GTK_WIDGET(gtk_builder_get_object(gtk_ppp.builder, "interface"));
	gtk_ppp.details.connect_speed = GTK_WIDGET(gtk_builder_get_object(gtk_ppp.builder, "connect_speed"));
	gtk_ppp.details.status = GTK_WIDGET(gtk_builder_get_object(gtk_ppp.builder, "status"));
	gtk_ppp.details.speed_in = GTK_WIDGET(gtk_builder_get_object(gtk_ppp.builder, "speed_in"));
	gtk_ppp.details.speed_out = GTK_WIDGET(gtk_builder_get_object(gtk_ppp.builder, "speed_out"));
	gtk_ppp.details.ip = GTK_WIDGET(gtk_builder_get_object(gtk_ppp.builder, "ip"));
	gtk_ppp.details.destination = GTK_WIDGET(gtk_builder_get_object(gtk_ppp.builder, "destination"));
	gtk_ppp.details.dns = GTK_WIDGET(gtk_builder_get_object(gtk_ppp.builder, "dns"));
	gtk_ppp.details.sent = GTK_WIDGET(gtk_builder_get_object(gtk_ppp.builder, "sent"));
	gtk_ppp.details.received = GTK_WIDGET(gtk_builder_get_object(gtk_ppp.builder, "received"));
	gtk_ppp.details.sent_packets = GTK_WIDGET(gtk_builder_get_object(gtk_ppp.builder, "sent_packets"));
	gtk_ppp.details.received_packets = GTK_WIDGET(gtk_builder_get_object(gtk_ppp.builder, "received_packets"));
	/* phone list window */
	gtk_ppp.phone.window = GTK_WIDGET(gtk_builder_get_object(gtk_ppp.builder, "phones"));
	gtk_window_set_icon_from_file (GTK_WINDOW (gtk_ppp.phone.window), icon, NULL);
	gtk_ppp.phone.phone_nums = GTK_WIDGET(gtk_builder_get_object(gtk_ppp.builder, "phone_nums"));
	/* init strings list window */
	gtk_ppp.init_strs.window = GTK_WIDGET(gtk_builder_get_object(gtk_ppp.builder, "init_strings"));
	gtk_window_set_icon_from_file (GTK_WINDOW (gtk_ppp.phone.window), icon, NULL);
	gtk_ppp.init_strs.init_strs = GTK_WIDGET(gtk_builder_get_object(gtk_ppp.builder, "init_strs"));

	/* free app icon path variable */
	g_free (icon);

	/* autoconnect callbacks */
	gtk_builder_connect_signals(gtk_ppp.builder, NULL);

	return gtk_ppp.gtk_ppp.window;
}


static void gtk_ppp_application_startup(GApplication *application, gpointer data)
{
	/*Create IP regular expression*/
	gtk_ppp.ipregex = g_regex_new(IPREGEX, G_REGEX_OPTIMIZE, G_REGEX_MATCH_NOTEMPTY, NULL);
}

static void gtk_ppp_application_activate(GApplication *application, gpointer data)
{
	GList *appwindows;
	GtkWidget *mainwin;

	appwindows = gtk_application_get_windows(GTK_APPLICATION(application));

	if (appwindows == NULL) {
		if (gtk_ppp_test_enviroment()) {
			/*Create application window and show it*/
			mainwin = gtk_ppp_create_main_window(GTK_APPLICATION(application));
			if (mainwin != NULL) {
				/* read ~/.wvdial.conf */
				read_conf ();
				gtk_application_add_window(GTK_APPLICATION(application), GTK_WINDOW(mainwin));
				gtk_widget_show(mainwin);
			} else {
				g_application_quit(application);
			}
		} else {
			g_application_quit(application);
		}
	} else {
		/*Windows already created, show main one*/
		gtk_window_present(GTK_WINDOW(appwindows->data));
	}
}

static void gtk_ppp_application_shutdown(GApplication *application, gpointer user_data)
{
	if (gtk_ppp.ipregex != NULL) {
		g_regex_unref(gtk_ppp.ipregex);
	}
}

int main (int argc, char *argv[])
{
	GtkApplication *app;
	gint res;
	GOptionContext *context;
	GError *error = NULL;
	GOptionEntry options [] =
	{ { "conf", 'c', 0, G_OPTION_ARG_STRING, &gtk_ppp.conf, N_("Wvdial configuration file"), N_("Use specified wvdial configuration file") },
      { NULL } };

	#ifdef ENABLE_NLS
	bindtextdomain (GETTEXT_PACKAGE, PACKAGE_LOCALE_DIR);
	bind_textdomain_codeset (GETTEXT_PACKAGE, "UTF-8");
	textdomain (GETTEXT_PACKAGE);
	#endif

	/* wvdial process */
	wvdial.cmd[0] = NULL;
	wvdial.out_pipe = 0;
	wvdial.err_pipe = 0;
	wvdial.in_pipe = 0;
	wvdial.pid = 0;
	wvdial.error = NULL;
	wvdial.out_tag = 0;
	wvdial.err_tag = 0;
	wvdial.out = NULL;
	wvdial.err = NULL;
	/* wvdialconf process */
	wvdialconf.cmd[0] = NULL;
	wvdialconf.out_pipe = 0;
	wvdialconf.err_pipe = 0;
	wvdialconf.in_pipe = 0;
	wvdialconf.pid = 0;
	wvdialconf.error = NULL;
	wvdialconf.out_tag = 0;
	wvdialconf.err_tag = 0;
	wvdialconf.out = NULL;
	wvdialconf.err = NULL;
	/* GtkBuilder xml */
	gtk_ppp.builder = NULL;
	/* gtk ppp window */
	gtk_ppp.gtk_ppp.window = NULL;
	gtk_ppp.gtk_ppp.username = NULL;
	gtk_ppp.gtk_ppp.password = NULL;
	gtk_ppp.gtk_ppp.store_pwd = NULL;
	gtk_ppp.gtk_ppp.phone = NULL;
	/* setup window */
	/* modem */
	gtk_ppp.setup.window = NULL;
	gtk_ppp.setup.device = NULL;
	gtk_ppp.setup.speed = NULL;
	gtk_ppp.setup.type = NULL;
	gtk_ppp.setup.tone = NULL;
	gtk_ppp.setup.pulse = NULL;
	gtk_ppp.setup.volume = NULL;
	gtk_ppp.setup.phone_nums = NULL;
	gtk_ppp.setup.init_strs = NULL;
	gtk_ppp.setup.dial_prefix = NULL;
	gtk_ppp.setup.spin_dial_attempts = NULL;
	gtk_ppp.setup.label_dial_attempts = NULL;
	gtk_ppp.setup.dial_tone = NULL;
	/* networking */
	gtk_ppp.setup.static_ip = NULL;
	gtk_ppp.setup.dynamic_ip = NULL;
	gtk_ppp.setup.label_ip = NULL;
	gtk_ppp.setup.entry_ip = NULL;
	gtk_ppp.setup.domain = NULL;
	gtk_ppp.setup.auto_dns = NULL;
	gtk_ppp.setup.manual_dns = NULL;
	gtk_ppp.setup.list_dns = NULL;
	/* options */
	gtk_ppp.setup.minimize = NULL;
	gtk_ppp.setup.dock = NULL;
	gtk_ppp.setup.auto_reconnect = NULL;
	gtk_ppp.setup.abort_on_busy = NULL;
	gtk_ppp.setup.abort_on_no_dialtone = NULL;
	gtk_ppp.setup.check_carrier = NULL;
	gtk_ppp.setup.check_def_route = NULL;
	gtk_ppp.setup.stupid_mode = NULL;
	gtk_ppp.setup.custom_reply = NULL;
	gtk_ppp.setup.label_reply = NULL;
	gtk_ppp.setup.entry_reply = NULL;
	gtk_ppp.setup.spin_idle = NULL;
	gtk_ppp.setup.label_idle = NULL;
	/* connecting window */
	gtk_ppp.connecting.window = NULL;
	gtk_ppp.connecting.msgs = NULL;
	/* log window */
	gtk_ppp.log.window = NULL;
	gtk_ppp.log.text_log = NULL;
	/* connected window */
	gtk_ppp.connected.window = NULL;
	gtk_ppp.connected.time = NULL;
	/* details window */
	gtk_ppp.details.window = NULL;
	gtk_ppp.details.interface = NULL;
	gtk_ppp.details.connect_speed = NULL;
	gtk_ppp.details.status = NULL;
	gtk_ppp.details.speed_in = NULL;
	gtk_ppp.details.speed_out = NULL;
	gtk_ppp.details.ip = NULL;
	gtk_ppp.details.destination = NULL;
	gtk_ppp.details.sent = NULL;
	gtk_ppp.details.received = NULL;
	gtk_ppp.details.sent_packets = NULL;
	gtk_ppp.details.received_packets = NULL;
	/* detect window */
	gtk_ppp.detect = NULL;
	/* notification */
	//gtk_ppp.notification.indicator = NULL;
	gtk_ppp.notification.icon = NULL;
	gtk_ppp.notification.mainimg = NULL;
	gtk_ppp.notification.sendingreceivingimg = NULL;
	gtk_ppp.notification.receivingimg = NULL;
	gtk_ppp.notification.sendingimg = NULL;
	gtk_ppp.notification.idleimg = NULL;
	gtk_ppp.notification.menu = NULL;
	/* about window */
	gtk_ppp.about = NULL;
	/* phone list window */
	gtk_ppp.phone.window = NULL;
	gtk_ppp.phone.phone_nums = NULL;
	/* init strins list window */
	gtk_ppp.init_strs.window = NULL;
	gtk_ppp.init_strs.init_strs = NULL;
	/* variables */
	gtk_ppp.connected.seconds = 0;
	gtk_ppp.connected.update = 0;
	gtk_ppp.connected.old_bytes_in = 0;
	gtk_ppp.connected.old_bytes_out = 0;
	gtk_ppp.connected.primarydns = NULL;
	gtk_ppp.connected.secondarydns = NULL;
	gtk_ppp.connected.show = TRUE;
	gtk_ppp.ipregex = NULL;
	gtk_ppp.dock = FALSE;
	/* conf path */
	gtk_ppp.conf = NULL;

	context = g_option_context_new(_("- GTK front-end to wvDial, a modem/ISDN dial-up software "));
	#ifdef ENABLE_NLS
	g_option_context_add_main_entries(context, options, NULL);
	#else
	g_option_context_add_main_entries(context, options, GETTEXT_PACKAGE);
	#endif

	if (!g_option_context_parse(context, &argc, &argv, &error)) {
		g_printf(_("Unable to parse command line arguments: %s\n"), error->message);
		g_option_context_free(context);
		g_error_free(error);
		exit(EXIT_FAILURE);
	}

	g_option_context_free(context);

	if (gtk_ppp.conf == NULL) {
		gtk_ppp.conf = g_strdup (".wvdial.conf");
	}

	app = gtk_application_new("org.gtk.GTK-PPP", G_APPLICATION_FLAGS_NONE);

	g_application_set_default(G_APPLICATION(app));

	g_signal_connect(app, "startup", G_CALLBACK(gtk_ppp_application_startup), NULL);
	g_signal_connect(app, "activate", G_CALLBACK(gtk_ppp_application_activate), NULL);
	g_signal_connect(app, "shutdown", G_CALLBACK(gtk_ppp_application_shutdown), NULL);

	res = g_application_run(G_APPLICATION(app), argc, argv);

	g_object_unref(app);

	return res;
}

