/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA.
 */

#include "gtk-ppp.h"

static gchar *extract_ip_address(gchar *string)
{
	gchar *eol, *addr;

	if (string == NULL) return NULL;

	/*Remove newline character*/
	eol = strrchr(string, '\n');
	if (eol != NULL) {
		eol[0] = '\0';
	}

	/*Get address*/
	addr = strrchr(string, ' ');
	if (addr == NULL) return NULL;
	if (addr[1] == '\0') return NULL;

	/*Use regular expression to validate address*/
	if (!g_regex_match_all(gtk_ppp.ipregex, addr+1, 0, NULL)) return NULL;

	return addr+1;
}

static gboolean resolv_conf_test_nameservers(gchar *nameserver1, gchar *nameserver2)
{
	FILE *resolvconf;
	gchar line[60];
	gchar *addr;
	gint checks;
	gboolean nsres, res;

	if ((nameserver1 == NULL) && (nameserver2 == NULL)) return FALSE;

	resolvconf = g_fopen("/etc/resolv.conf", "r");

	if (resolvconf == NULL) return FALSE;

	res = TRUE;
	checks = 0;

	while (fgets(line, sizeof(line), resolvconf) != NULL) {
		if (strstr(line, "nameserver") != NULL) {
			addr = extract_ip_address(line);
			if (addr != NULL) {
				nsres = FALSE;
				checks++;
				if (nameserver1 != NULL) {
					nsres = g_str_equal(nameserver1, addr);
				}
				if ((!nsres) && (nameserver2 != NULL)) {
					nsres = g_str_equal(nameserver2, addr);
				}
				if (!nsres) {
					res = FALSE;
					break;
				}
			}
		}
	}

	fclose(resolvconf);

	return res && (checks > 0);
}

static gboolean resolv_conf_update_nameservers(gchar *nameserver1, gchar *nameserver2)
{
	FILE *resolvconf;
	#ifdef ENABLE_SELINUX
	security_context_t secontext;
	#endif
	GString *resolvstr;
	gchar line[60];
	gboolean res;

	if ((nameserver1 == NULL) && (nameserver2 == NULL)) return FALSE;

	#ifdef ENABLE_SELINUX
	secontext = NULL;
	#endif

	resolvstr = g_string_new(NULL);

	resolvconf = g_fopen("/etc/resolv.conf", "r");

	if (resolvconf != NULL) {
		/*Copy all existing lines skipping nameservers*/
		while (fgets(line, sizeof(line), resolvconf) != NULL) {
			if (strstr(line, "nameserver") == NULL) {
				g_string_append(resolvstr, line);
			}
		}
		fclose(resolvconf);

		#ifdef ENABLE_SELINUX
		/*Save SELinux context*/
		if (getfilecon("/etc/resolv.conf", &secontext) == -1) {
			g_warning("Unable to save resolv.conf file SELinux context");
		}
		#endif
	}

	/*Then write updated nameservers addresses*/
	if (nameserver1 != NULL) {
		g_string_append_printf(resolvstr, "nameserver %s\n", nameserver1);
	}
	if (nameserver2 != NULL) {
		g_string_append_printf(resolvstr, "nameserver %s\n", nameserver2);
	}

	/*Update file contents*/
	res = g_file_set_contents("/etc/resolv.conf", resolvstr->str, resolvstr->len, NULL);
	if (res) {
		g_printf("Nameservers entries in resolv.conf file were updated after 5 seconds\n");
	}
	g_string_free(resolvstr, TRUE);

	#ifdef ENABLE_SELINUX
	/*Restore SELinux context*/
	if (secontext != NULL) {
		if (setfilecon("/etc/resolv.conf", secontext) == -1) {
			g_warning("Unable to restore resolv.conf file SELinux context");
		}
		freecon(secontext);
	}
	#endif

	return res;
}

static gboolean resolv_conf_update (gpointer data)
{

	if (!resolv_conf_test_nameservers(gtk_ppp.connected.primarydns, gtk_ppp.connected.secondarydns)) {
		resolv_conf_update_nameservers(gtk_ppp.connected.primarydns, gtk_ppp.connected.secondarydns);
	} else {
		g_printf("Nameservers entries in resolv.conf file are up to date\n");
	}

	return G_SOURCE_REMOVE;
}

gboolean on_wvdial_out (GIOChannel *source, GIOCondition condition, gpointer data)
{
	gchar *line;
	GIOStatus status;

	status = g_io_channel_read_line (source, &line, NULL, NULL, NULL);

	if (status != G_IO_STATUS_NORMAL)
		return FALSE;

	g_print ("GTK PPP: STDOUT: %s", line);
	g_free (line);

	return TRUE;
}

gboolean on_wvdial_err (GIOChannel *source, GIOCondition condition, gpointer data)
{
	gchar *line;
	GIOStatus status;
	GIOChannel *in; /* stdin */
	GtkTextBuffer *txt;
	GtkTextIter end_iter;

	status = g_io_channel_read_line (source, &line, NULL, NULL, NULL);

	if (status != G_IO_STATUS_NORMAL)
		return FALSE;

	g_print ("GTK PPP: STDERR: %s", line);

	if (g_strrstr (line, "--> Cannot open") != NULL)
		gtk_label_set_text (GTK_LABEL (gtk_ppp.connecting.msgs), _("Cannot open modem."));

	if (g_strrstr (line, "--> Modem not responding") != NULL)
		gtk_label_set_text (GTK_LABEL (gtk_ppp.connecting.msgs), _("Modem not responding."));

	if (g_strrstr (line, "--> Please enter password") != NULL) {
		const gchar *password;

		gtk_label_set_text (GTK_LABEL (gtk_ppp.connecting.msgs), _("Sending password..."));

		password = gtk_entry_get_text (GTK_ENTRY (gtk_ppp.gtk_ppp.password));

		in = g_io_channel_unix_new (wvdial.in_pipe);
		g_io_channel_set_encoding (in, NULL, NULL);
		g_io_channel_write_chars (in, password, -1, NULL, NULL);
		g_io_channel_shutdown (in, TRUE, NULL);
	}

	if (g_strrstr (line, "--> Configuration does not specify a valid password.") != NULL)
		gtk_label_set_text (GTK_LABEL (gtk_ppp.connecting.msgs), _("No valid password."));

	if (g_strrstr (line, "ATM") != NULL) {
		const gchar *phone;
		gchar *message;

		phone = gtk_entry_get_text(GTK_ENTRY(gtk_ppp.gtk_ppp.phoneentry));
		message = g_strdup_printf (_("Dialing %s..."), phone);

		gtk_label_set_text (GTK_LABEL (gtk_ppp.connecting.msgs), message);

		g_free (message);
	}

	if (g_strrstr (line, "--> Carrier detected.  Waiting for prompt.") != NULL)
		gtk_label_set_text (GTK_LABEL (gtk_ppp.connecting.msgs), _("Waiting for prompt..."));

	if (g_strrstr (line, "Username:") != NULL)
		gtk_label_set_text (GTK_LABEL (gtk_ppp.connecting.msgs), _("Authenticating..."));

	if (g_strrstr (line, "--> Using interface") != NULL) {
		gchar iface[255];
		gchar tmp[255];

		sscanf (line, "%s %s %s %s", tmp, tmp, tmp, iface);
		gtk_label_set_text (GTK_LABEL (gtk_ppp.details.interface), &iface[0]);
	}
	if (g_strrstr (line, "CONNECT ") != NULL) {
		gchar connect_speed[255];
		gchar tmp[255];
		sscanf (line, "%254s %254s", tmp, connect_speed);
		gtk_label_set_text (GTK_LABEL (gtk_ppp.details.connect_speed), &connect_speed[0]);
	}
	/* in case we are using auto reconnect */
	if (g_strrstr (line, "--> Auto Reconnect") != NULL) {
		gtk_label_set_text (GTK_LABEL (gtk_ppp.connecting.msgs), _("Reconnecting..."));
		gtk_widget_hide (gtk_ppp.gtk_ppp.window);
		gtk_widget_hide (gtk_ppp.details.window);
		gtk_widget_show (gtk_ppp.connecting.window);
	}
	/* we're about to connect (pppd) */
	if (g_strrstr (line, "--> Starting pppd") != NULL) {
		gboolean minimize;

		minimize = gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (gtk_ppp.setup.minimize));

		gtk_label_set_text (GTK_LABEL (gtk_ppp.connecting.msgs), _("Starting pppd..."));
		/* show connected window */
		gtk_widget_hide (gtk_ppp.connecting.window);
		//gtk_widget_hide (gtk_ppp.log.window);
		gtk_widget_show (gtk_ppp.connected.window);

		if (minimize)
			gtk_window_iconify (GTK_WINDOW (gtk_ppp.connected.window));

		if (gtk_ppp.dock == TRUE) {
			init_notification ();
			update_notification (3);
			show_notification (TRUE);
			gtk_widget_hide (gtk_ppp.connected.window);
		}
		/* start timer */
		gtk_ppp.connected.update = g_timeout_add (1000, (GSourceFunc)timer, NULL);
	}

	if (g_strrstr (line, "--> The PPP daemon has died") != NULL
		|| g_strrstr (line, "--> Disconnecting") != NULL) {

		if (gtk_ppp.connected.update > 0) {
			g_source_remove (gtk_ppp.connected.update);
			gtk_ppp.connected.update = 0;
			gtk_ppp.connected.seconds = 0;
			gtk_ppp.connected.old_bytes_in = 0;
			gtk_ppp.connected.old_bytes_out = 0;
			if (gtk_ppp.connected.primarydns != NULL) {
				g_free(gtk_ppp.connected.primarydns);
				gtk_ppp.connected.primarydns = NULL;
			}
			if (gtk_ppp.connected.secondarydns != NULL) {
				g_free(gtk_ppp.connected.secondarydns);
				gtk_ppp.connected.secondarydns = NULL;
			}

			gtk_label_set_text (GTK_LABEL (gtk_ppp.details.dns), _("Unknown"));

			gtk_label_set_markup (GTK_LABEL (gtk_ppp.connected.time),
						_("<big>Connected: 00:00:00</big>"));

			gtk_window_set_title (GTK_WINDOW(gtk_ppp.connected.window), "00:00:00");
		}

		if (gtk_ppp.notification.icon != NULL) {
			show_notification (FALSE);
			/* destroy notification */
			destroy_notification ();
		}
		gtk_widget_hide (gtk_ppp.connected.window);
		gtk_widget_hide (gtk_ppp.details.window);
		/* if there was an auto reconnect */
		gtk_widget_hide (gtk_ppp.connecting.window);
		gtk_widget_show (gtk_ppp.gtk_ppp.window);
	}

	/*Primary DNS address*/
	if (g_strrstr (line, "primary   DNS address") != NULL) {
		gchar *dns, *dnsstr;
		dns = extract_ip_address(line);
		if (dns != NULL) {
			gtk_ppp.connected.primarydns = g_strdup(dns);
			if (gtk_ppp.connected.secondarydns != NULL) {
				/*Show in details window*/
				dnsstr = g_strdup_printf("%s, %s", gtk_ppp.connected.primarydns, gtk_ppp.connected.secondarydns);
				gtk_label_set_text(GTK_LABEL(gtk_ppp.details.dns), dnsstr);
				g_free(dnsstr);
				/*Update resolv.conf if possible*/
				if (g_access("/etc", W_OK) != -1) {
					g_timeout_add(5000, resolv_conf_update, NULL);
				}
			}
		}
	}

	/*Secondary DNS address*/
	if (g_strrstr (line, "secondary DNS address") != NULL) {
		gchar *dns, *dnsstr;
		dns = extract_ip_address(line);
		if (dns != NULL) {
			gtk_ppp.connected.secondarydns = g_strdup(dns);
			if (gtk_ppp.connected.primarydns != NULL) {
				/*Show in details window*/
				dnsstr = g_strdup_printf("%s, %s", gtk_ppp.connected.primarydns, gtk_ppp.connected.secondarydns);
				gtk_label_set_text(GTK_LABEL(gtk_ppp.details.dns), dnsstr);
				g_free(dnsstr);
				/*Update resolv.conf if possible*/
				if (g_access("/etc", W_OK) != -1) {
					g_timeout_add(5000, resolv_conf_update, NULL);
				}
			}
		}
	}

	if (g_utf8_validate(line, -1, NULL)) {
		/* write wvdial output */
		txt = gtk_text_view_get_buffer (GTK_TEXT_VIEW (gtk_ppp.log.text_log));
		gtk_text_buffer_get_end_iter (txt, &end_iter);
		gtk_text_buffer_insert (txt, &end_iter, line, -1);
		/* scroll text to end */
		gtk_text_buffer_get_end_iter (txt, &end_iter);
		gtk_text_buffer_move_mark_by_name (txt, "insert", &end_iter);
		gtk_text_view_scroll_to_mark (GTK_TEXT_VIEW (gtk_ppp.log.text_log),
										gtk_text_buffer_get_mark (txt, "insert"),
										0.0, FALSE, 0.0, 0.0);
	}

	g_free (line);

	return TRUE;
}
